/*
 * Segment 10 System (Sys), SWE 10.1 StartStopp
 * Copyright (C) 2007-2017 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:<br>
 * BitCtrl Systems GmbH<br>
 * Weißenfelser Straße 67<br>
 * 04229 Leipzig<br>
 * Phone: +49 341-490670<br>
 * mailto: info@bitctrl.de
 */

package de.bsvrz.sys.startstopp.config;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import de.bsvrz.dav.daf.util.cron.CronDefinition;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.startstopp.api.StartStoppException;
import de.bsvrz.sys.startstopp.api.jsonschema.Global;
import de.bsvrz.sys.startstopp.api.jsonschema.Inkarnation;
import de.bsvrz.sys.startstopp.api.jsonschema.KernSystem;
import de.bsvrz.sys.startstopp.api.jsonschema.MakroDefinition;
import de.bsvrz.sys.startstopp.api.jsonschema.Rechner;
import de.bsvrz.sys.startstopp.api.jsonschema.StartArt;
import de.bsvrz.sys.startstopp.api.jsonschema.StartBedingung;
import de.bsvrz.sys.startstopp.api.jsonschema.StartFehlerVerhalten;
import de.bsvrz.sys.startstopp.api.jsonschema.StartStoppSkript;
import de.bsvrz.sys.startstopp.api.jsonschema.StartStoppSkriptStatus;
import de.bsvrz.sys.startstopp.api.jsonschema.StartStoppSkriptStatus.Status;
import de.bsvrz.sys.startstopp.api.jsonschema.StoppBedingung;
import de.bsvrz.sys.startstopp.api.jsonschema.StoppFehlerVerhalten;
import de.bsvrz.sys.startstopp.api.jsonschema.Usv;
import de.bsvrz.sys.startstopp.api.util.Util;
import de.bsvrz.sys.startstopp.process.OnlineInkarnation;

public final class StartStoppKonfiguration {

    private static final Debug LOGGER = Debug.getLogger();
    private StartStoppSkript skript;
    private StartStoppSkriptStatus skriptStatus = new StartStoppSkriptStatus();
    private String checkSumme = "";

    public static void fuelleStandardWerte(StartStoppSkript konfiguration) {
        for (Inkarnation inkarnation : konfiguration.getInkarnationen()) {
            StartArt startArt = inkarnation.getStartArt();
            if (startArt == null) {
                inkarnation.setStartArt(new StartArt());
            }
            StartFehlerVerhalten startFehlerVerhalten = inkarnation.getStartFehlerVerhalten();
            if (startFehlerVerhalten == null) {
                inkarnation.setStartFehlerVerhalten(new StartFehlerVerhalten());
            }
            StoppFehlerVerhalten stoppFehlerVerhalten = inkarnation.getStoppFehlerVerhalten();
            if (stoppFehlerVerhalten == null) {
                inkarnation.setStoppFehlerVerhalten(new StoppFehlerVerhalten());
            }
        }
    }

    public StartStoppKonfiguration(StartStoppSkript skript) {
        this.skript = skript;
        fuelleStandardWerte(skript);

        skriptStatus.getMessages().addAll(pruefeVollstaendigkeit());
        skriptStatus.getMessages().addAll(pruefeZirkularitaet());

        if (skriptStatus.getMessages().isEmpty()) {
            skriptStatus.setStatus(StartStoppSkriptStatus.Status.INITIALIZED);
        } else {
            skriptStatus.setStatus(StartStoppSkriptStatus.Status.FAILURE);
        }
    }

    private Collection<String> pruefeZirkularitaet() {

        Collection<String> result = new ArrayList<>();

        try {
            getResolvedMakros();
        } catch (StartStoppException e) {
            result.add(e.getLocalizedMessage());
        }

        for (Inkarnation inkarnation : skript.getInkarnationen()) {
            try {
                checkStartRules(inkarnation);
            } catch (StartStoppException e) {
                result.add(e.getLocalizedMessage());
            }
            try {
                checkStopRules(inkarnation);
            } catch (StartStoppException e) {
                result.add(e.getLocalizedMessage());
            }
        }
        return result;
    }

    private OnlineInkarnation getInkarnation(String name) throws StartStoppException {

        if (name == null) {
            throw new StartStoppException("Inkarnationsname fehlt!");
        }

        for (Inkarnation inkarnation : skript.getInkarnationen()) {
            if (name.equals(inkarnation.getInkarnationsName())) {
                return new OnlineInkarnation(this, inkarnation);
            }
        }
        throw new StartStoppException("Ein Inkarnation \"" + name + "\" ist nicht definiert");
    }

    private void checkStartRules(Inkarnation inkarnation) throws StartStoppException {

        Set<String> usedInkarnations = new LinkedHashSet<>();
        Queue<Inkarnation> mustBeChecked = new ConcurrentLinkedQueue<>();
        mustBeChecked.add(inkarnation);

        while (!mustBeChecked.isEmpty()) {

            Inkarnation currentInkarnation = mustBeChecked.poll();
            if (!usedInkarnations.add(currentInkarnation.getInkarnationsName())) {
                throw new StartStoppException(
                        "Regeln für \"" + inkarnation.getInkarnationsName() + "\" sind rekursiv!");
            }

            StartBedingung startBedingung = currentInkarnation.getStartBedingung();
            if (startBedingung == null) {
                continue;
            }
            String rechnerName = startBedingung.getRechner();
            if (rechnerName != null && !rechnerName.trim().isEmpty()) {
                checkRechner(currentInkarnation, rechnerName.trim());
                continue;
            }
            try {
                for (String vorgaenger : startBedingung.getVorgaenger()) {
                    mustBeChecked.add(getInkarnation(vorgaenger).getInkarnation());
                }
            } catch (StartStoppException e) {
                LOGGER.fine(e.getLocalizedMessage());
                throw new StartStoppException(
                        currentInkarnation.getInkarnationsName() + ": StartRegel hat keinen Vorgänger!");
            }
        }
    }

    private void checkRechner(Inkarnation inkarnation, String name) throws StartStoppException {
        for (Rechner rechner : getResolvedRechner()) {
            if (rechner.getName().equals(name)) {
                return;
            }
        }

        throw new StartStoppException("Der in der Inkarnation \"" + inkarnation.getInkarnationsName()
                + "\" referenzierte Rechner \"" + name + "\" ist nicht in der Konfiguration definiert");
    }

    private void checkStopRules(Inkarnation inkarnation) throws StartStoppException {

        Set<String> usedInkarnations = new LinkedHashSet<>();
        Queue<Inkarnation> mustBeChecked = new ConcurrentLinkedQueue<>();
        mustBeChecked.add(inkarnation);

        while (!mustBeChecked.isEmpty()) {

            Inkarnation currentInkarnation = mustBeChecked.poll();
            if (!usedInkarnations.add(currentInkarnation.getInkarnationsName())) {
                throw new StartStoppException(
                        "Regeln für \"" + inkarnation.getInkarnationsName() + "\" sind rekursiv!");
            }

            StoppBedingung stoppBedingung = currentInkarnation.getStoppBedingung();
            if (stoppBedingung == null) {
                continue;
            }
            String rechner = stoppBedingung.getRechner();
            if (rechner != null && !rechner.trim().isEmpty()) {
                checkRechner(currentInkarnation, rechner.trim());
                continue;
            }
            try {
                for (String nachfolger : stoppBedingung.getNachfolger()) {
                    mustBeChecked.add(getInkarnation(nachfolger).getInkarnation());
                }
            } catch (StartStoppException e) {
                LOGGER.fine(e.getLocalizedMessage());
                throw new StartStoppException(
                        currentInkarnation.getInkarnationsName() + ": StopRegel hat keinen Nachfolger!");
            }
        }
    }

    @SuppressWarnings("unused")
    private Collection<String> pruefeVollstaendigkeit() {

        Collection<String> result = new ArrayList<>();

        /* Prüfung der Datenverteilerzugangsdaten. */

        try {
            getResolvedRechner();
        } catch (StartStoppException e) {
            result.add(e.getLocalizedMessage());
        }

        for (KernSystem kernSystem : getKernSysteme()) {
            try {
                getInkarnation(kernSystem.getInkarnationsName());
            } catch (StartStoppException e) {
                result.add("Kernsystem: " + e.getLocalizedMessage());
            }
        }

        for (Inkarnation inkarnation : skript.getInkarnationen()) {
            try {
                OnlineInkarnation onlineInkarnation = new OnlineInkarnation(this, inkarnation);
                try {
                    if (onlineInkarnation.getInkarnation().getStartBedingung() != null) {
                        Util.convertToWarteZeitInMsec(
                                onlineInkarnation.getInkarnation().getStartBedingung().getWartezeit());
                    }
                } catch (StartStoppException e) {
                    result.add(onlineInkarnation.getInkarnation().getInkarnationsName()
                            + ": Wartezeit für die Startbedingung kann nicht interpretiert werden: "
                            + e.getLocalizedMessage());
                }
                try {
                    if (onlineInkarnation.getInkarnation().getStoppBedingung() != null) {
                        Util.convertToWarteZeitInMsec(
                                onlineInkarnation.getInkarnation().getStoppBedingung().getWartezeit());
                    }
                } catch (StartStoppException e) {
                    result.add(onlineInkarnation.getInkarnation().getInkarnationsName()
                            + ": Wartezeit für die Stoppbedingung kann nicht interpretiert werden: "
                            + e.getLocalizedMessage());
                }
                switch (onlineInkarnation.getInkarnation().getStartArt().getOption()) {
                case INTERVALLABSOLUT:
                    try {
                        new CronDefinition(onlineInkarnation.getInkarnation().getStartArt().getIntervall());
                    } catch (IllegalArgumentException e) {
                        result.add(onlineInkarnation.getInkarnation().getInkarnationsName()
                                + ": Die CRON-Syntax für die zyklische Ausführung kann nicht interpretiert werden: "
                                + e.getLocalizedMessage());
                    }
                    break;
                case INTERVALLRELATIV:
                    try {
                        Util.convertToWarteZeitInMsec(onlineInkarnation.getInkarnation().getStartArt().getIntervall());
                    } catch (StartStoppException e) {
                        result.add(onlineInkarnation.getInkarnation().getInkarnationsName()
                                + ": Das Intervall für die zyklische Ausführung kann nicht interpretiert werden: "
                                + e.getLocalizedMessage());
                    }
                    break;
                case MANUELL:
                case AUTOMATISCH:
                default:
                    break;

                }
            } catch (StartStoppException e) {
                result.add(e.getLocalizedMessage());
            }
        }

        return result;
    }

    public StartStoppSkript getSkript() {
        return skript;
    }

    public StartStoppSkriptStatus getSkriptStatus() {
        return skriptStatus;
    }

    public Collection<OnlineInkarnation> getInkarnationen() throws StartStoppException {
        if (skriptStatus.getStatus() != StartStoppSkriptStatus.Status.INITIALIZED) {
            throw new StartStoppException("Das geladene StartStoppSkript ist nicht korrekt versioniert!");
        }

        Collection<OnlineInkarnation> result = new ArrayList<>();
        for (Inkarnation inkarnation : skript.getInkarnationen()) {
            OnlineInkarnation startStoppInkarnation = new OnlineInkarnation(this, inkarnation);
            result.add(startStoppInkarnation);
        }
        return result;
    }

    public Map<String, String> getResolvedMakros() throws StartStoppException {

        Pattern pattern = Pattern.compile("%.*?%");
        Map<String, String> result = new LinkedHashMap<>();

        Global global = skript.getGlobal();
        if (global != null) {
            for (MakroDefinition makroDefinition : global.getMakrodefinitionen()) {
                String name = makroDefinition.getName();
                String wert = makroDefinition.getWert();
                Set<String> usedMakros = new LinkedHashSet<>();

                do {
                    boolean replaced = false;
                    Matcher matcher = pattern.matcher(wert);
                    while (matcher.find()) {
                        String part = matcher.group();
                        String key = part.substring(1, part.length() - 1);
                        String replacement = result.get(key);
                        if (replacement == null) {
                            if (!usedMakros.add(key)) {
                                throw new StartStoppException(
                                        "Makros können wegen einer Rekursion nicht aufgelöst werden!");
                            }
                            replacement = getMakroValueFor(key);
                        }
                        wert = wert.replaceAll(part, replacement);
                        replaced = true;
                    }
                    if (!replaced) {
                        result.put(name, wert);
                        break;
                    }
                } while (true);
            }
        }

        return result;
    }

    private String getMakroValueFor(String key) throws StartStoppException {
        for (MakroDefinition makroDefinition : skript.getGlobal().getMakrodefinitionen()) {
            if (key.equals(makroDefinition.getName())) {
                return makroDefinition.getWert();
            }
        }

        throw new StartStoppException("Für das Makro \"" + key + "\" wurde kein Wert definiert");
    }

    public Collection<Rechner> getResolvedRechner() throws StartStoppException {

        Map<String, String> resolvedMakros = getResolvedMakros();
        Pattern pattern = Pattern.compile("%.*?%");
        Collection<Rechner> result = new ArrayList<>();

        Global global = getSkript().getGlobal();
        if (global != null) {
            for (Rechner rechner : global.getRechner()) {
                Rechner resolvedRechner = new Rechner();
                resolvedRechner.setName(rechner.getName());

                String adresse = rechner.getTcpAdresse();
                Matcher matcher = pattern.matcher(adresse);
                while (matcher.find()) {
                    String part = matcher.group();
                    String key = part.substring(1, part.length() - 1);
                    String replacement = resolvedMakros.get(key);
                    if (replacement == null) {
                        throw new StartStoppException("Das Makro \"" + key + "\" für den Rechner \"" + rechner.getName()
                                + "\" ist nicht definiert!");
                    }
                    adresse = adresse.replaceAll(part, replacement);
                }
                resolvedRechner.setTcpAdresse(adresse);

                String port = rechner.getPort();
                matcher = pattern.matcher(port);
                while (matcher.find()) {
                    String part = matcher.group();
                    String key = part.substring(1, part.length() - 1);
                    String replacement = resolvedMakros.get(key);
                    if (replacement == null) {
                        throw new StartStoppException("Das Makro \"" + key + "\" für den Rechner \"" + rechner.getName()
                                + "\" ist nicht definiert!");
                    }
                    port = port.replaceAll(part, replacement);
                }
                resolvedRechner.setPort(port);

                result.add(resolvedRechner);
            }
        }
        return result;
    }

    public Rechner getResolvedRechner(String rechnerName) throws StartStoppException {

        for (Rechner rechner : getResolvedRechner()) {
            if (rechner.getName().equals(rechnerName)) {
                return rechner;
            }
        }

        throw new StartStoppException("Ein Rechner mit dem Name \"" + rechnerName + "\" ist nicht definiert!");
    }

    public String makroResolvedString(String quelle, String wert) throws StartStoppException {

        if (wert == null) {
            return null;
        }

        String result = wert;

        Map<String, String> resolvedMakros = getResolvedMakros();
        Pattern pattern = Pattern.compile("%.*?%");
        Matcher matcher = pattern.matcher(result);
        while (matcher.find()) {
            String part = matcher.group();
            String key = part.substring(1, part.length() - 1);
            String replacement = resolvedMakros.get(key);
            if (replacement == null) {
                throw new StartStoppException("Das Makro \"" + key + "\" für " + quelle + " ist nicht definiert!");
            }
            result = result.replaceAll(part, replacement);
        }

        return result;
    }

    public StartBedingung getResolvedStartBedingung(StartBedingung startBedingung) throws StartStoppException {

        if (startBedingung == null) {
            return null;
        }

        StartBedingung bedingung = new StartBedingung();
        bedingung.setVorgaenger(startBedingung.getVorgaenger());
        bedingung.setWarteart(startBedingung.getWarteart());
        bedingung.setRechner(startBedingung.getRechner());
        bedingung.setWartezeit(makroResolvedString("Startbedingung", startBedingung.getWartezeit()));
        return bedingung;
    }

    public StartFehlerVerhalten getResolvedStartFehlerVerhalten(StartFehlerVerhalten startFehlerVerhalten) {

        if (startFehlerVerhalten == null) {
            return null;
        }

        StartFehlerVerhalten verhalten = new StartFehlerVerhalten();
        verhalten.setOption(startFehlerVerhalten.getOption());
        verhalten.setWiederholungen(startFehlerVerhalten.getWiederholungen());
        return verhalten;
    }

    public StoppBedingung getResolvedStoppBedingung(StoppBedingung stoppBedingung) throws StartStoppException {

        if (stoppBedingung == null) {
            return null;
        }

        StoppBedingung bedingung = new StoppBedingung();
        bedingung.setNachfolger(stoppBedingung.getNachfolger());
        bedingung.setRechner(stoppBedingung.getRechner());
        bedingung.setWartezeit(makroResolvedString("StoppBedingung", stoppBedingung.getWartezeit()));
        return bedingung;
    }

    public StoppFehlerVerhalten getResolvedStoppFehlerVerhalten(StoppFehlerVerhalten stoppFehlerVerhalten) {

        if (stoppFehlerVerhalten == null) {
            return null;
        }

        StoppFehlerVerhalten verhalten = new StoppFehlerVerhalten();
        verhalten.setOption(stoppFehlerVerhalten.getOption());
        verhalten.setWiederholungen(stoppFehlerVerhalten.getWiederholungen());
        return verhalten;
    }

    public void setSkriptStatus(Status status, String message) {
        skriptStatus.setStatus(status);
        skriptStatus.getMessages().add(message);
    }

    public String getCheckSumme() {
        return checkSumme;
    }

    public void setCheckSumme(String checkSumme) {
        this.checkSumme = checkSumme;
    }

    public List<KernSystem> getKernSysteme() {
        List<KernSystem> result = new ArrayList<>();
        if (skript.getGlobal() != null) {
            result.addAll(skript.getGlobal().getKernsysteme());
        }
        return result;
    }

    public Usv getResolvedUsv() throws StartStoppException {

        Map<String, String> resolvedMakros = getResolvedMakros();
        Pattern pattern = Pattern.compile("%.*?%");

        Usv usv = getSkript().getGlobal().getUsv();
        if (usv == null) {
            return null;
        }

        Usv resolvedUsv = new Usv();
        String pid = usv.getPid();
        Matcher matcher = pattern.matcher(pid);
        while (matcher.find()) {
            String part = matcher.group();
            String key = part.substring(1, part.length() - 1);
            String replacement = resolvedMakros.get(key);
            if (replacement == null) {
                throw new StartStoppException("Das Makro \"" + key + "\" ist für die USV nicht definiert!");
            }
            pid = pid.replaceAll(part, replacement);
        }
        resolvedUsv.setPid(pid);
        return resolvedUsv;
    }
}
